#include "CompileStep.h"

#include <fstream>

#include "sys.h"
#include "toolchain.h"
#include "logger.h"
#include "globals.h"

static bool startsWith(BBParser &parser, const char *str) {
	for (size_t i = 0; str[i] != '\0'; ++i) {
		if (parser.peek() != str[i])
			return false;
		parser.skip();
	}

	return true;
}

bool CompileStep::checkHasChanged(const std::string &outDir) {
	std::string objPath = toolchain::objectFilePath(path_, outDir);
	if (!sys::fileExists(objPath)) {
		return true;
	}

	sys::FileInfo objInfo = sys::fileInfo(objPath);
	sys::FileInfo sourceInfo = sys::fileInfo(path_);
	if (objInfo.isOlderThan(sourceInfo)) {
		return true;
	}

	std::string bbPath = this->bbPath(outDir);
	if (!sys::fileExists(bbPath)) {
		return true;
	}

	BBVariables cachedVariables;
	try {
		std::ifstream f(bbPath);
		BBParser parser(f, BBParser::FLAG_NONE);
		parser.parse(cachedVariables);
	} catch (BBParseError &err) {
		logger::log(bbPath + ": " + err.what());
		return true;
	}

	// We can trust the cached "deps" value here, because we know
	// the object file hasn't actually changed.
	auto depsIt = cachedVariables.find("deps");
	if (depsIt == cachedVariables.end()) {
		logger::log(bbPath + ": Missing 'deps' field");
		return true;
	}

	const std::vector<std::string> &deps = depsIt->second;
	for (const auto &dep: deps) {
		if (!sys::fileExists(dep)) {
			return true;
		}

		sys::FileInfo depInfo = sys::fileInfo(dep);
		if (objInfo.isOlderThan(depInfo)) {
			return true;
		}
	}

	// Maybe the build command has changed?
	auto commandIt = cachedVariables.find("command");
	if (commandIt == cachedVariables.end()) {
		logger::log(bbPath + ": Missing 'command' field");
		return true;
	}

	if (compileCommand(outDir) != commandIt->second) {
		return true;
	}

	return false;
}

void CompileStep::doBuild(const std::string &outDir) {
	logger::log("Compile " + path_);

	std::string objPath = toolchain::objectFilePath(path_, outDir);
	std::string dirPath = sys::dirname(objPath);
	std::vector<std::string> command = compileCommand(outDir);

	BBVariables newCachedVars = {
		{ "deps", toolchain::getDependencies(flags(), type_, path_) },
		{ "command", command},
	};

	std::ofstream f(bbPath(outDir));
	BBWriter writer(f);
	writer.write(newCachedVars);

	sys::mkdirp(dirPath);
	sys::execute(command, nullptr, global::verbose >= 1);
}

std::vector<std::string> CompileStep::getPublicLDFlags(const std::string &outDir) {
	BBVariables &vars = variables();
	auto it = vars.find("ldflags");
	if (it == vars.end()) {
		return {};
	} else {
		return it->second;
	}
}

std::vector<std::string> CompileStep::getPublicLDLibs(const std::string &outDir) {
	std::vector<std::string> libs;
	BBVariables &vars = variables();

	auto pkgsIt = vars.find("pkgs");
	if (pkgsIt != vars.end()) {
		toolchain::getPkgConfigLDLibs(pkgsIt->second, libs);
	}

	auto libsIt = vars.find("ldlibs");
	if (libsIt != vars.end()) {
		for (auto &lib: libsIt->second) {
			libs.push_back(lib);
		}
	}

	return libs;
}

std::vector<std::string> CompileStep::getPublicObjects(const std::string &outDir) {
	return std::vector<std::string>{ toolchain::objectFilePath(path_, outDir) };
}

BBVariables &CompileStep::variables() {
	if (hasVariables_) {
		return variables_;
	}

	std::ifstream f(path_);
	BBParser parser(f, BBParser::FLAG_ONE_LINE);

	while (f.good()) {
		if (startsWith(parser, "//#bb")) {
			parser.parse(variables_);
		} else {
			while (f.good() && parser.get() != '\n');
		}
	}

	hasVariables_ = true;
	return variables_;
}

std::vector<std::string> &CompileStep::flags() {
	if (hasFlags_) {
		return flags_;
	}

	BBVariables &vars = variables();
	std::vector<std::string> flags;

	auto pkgsIt = vars.find("pkgs");
	if (pkgsIt != vars.end()) {
		for (auto &flag: pkgsIt->second) {
			flags.push_back(flag);
		}
	}

	std::string cflagsName;
	switch (type_) {
		case toolchain::FileType::C:
			cflagsName = "cflags";
			break;
		case toolchain::FileType::CXX:
			cflagsName = "cxxflags";
			break;
	}

	auto cflagsIt = vars.find(cflagsName);
	if (cflagsIt != vars.end()) {
		for (auto &flag: cflagsIt->second) {
			flags.push_back(flag);
		}
	}

	hasFlags_ = true;
	return flags_;
}

std::vector<std::string> &CompileStep::compileCommand(const std::string &outDir) {
	if (hasCompileCommand_) {
		return compileCommand_;
	}

	compileCommand_ = toolchain::getCompileCommand(flags(), type_, path_, outDir);
	hasCompileCommand_ = true;
	return compileCommand_;
}
