#include "build.h"

#include <fstream>
#include <stdexcept>
#include <algorithm>
#include <string.h>

#include "sys.h"
#include "BBParser.h"
#include "CompileStep.h"
#include "LinkStep.h"

static std::string extension(const std::string &path) {
	size_t idx = path.find_last_of('.');
	if (idx >= std::string::npos) {
		return "";
	}

	char ext[16] = { 0 };
	strncpy(ext, path.c_str() + idx, sizeof(ext) - 1);
	return ext;
}

static void addFile(
		std::string path, BBVariables variables,
		std::vector<std::unique_ptr<DepNode>> &deps) {
	toolchain::FileType type;
	std::string ext = extension(path);
	if (ext == ".c") {
		type = toolchain::FileType::C;
	} else if (ext == ".C" || ext == ".cc" || ext == ".cxx" || ext == ".cpp") {
		type = toolchain::FileType::CXX;
	} else {
		return;
	}

	deps.push_back(std::make_unique<CompileStep>(std::move(path), type, std::move(variables)));
}

static void findDeps(
		const std::string &dir, const std::string &name,
		const BBVariables &variables, std::vector<std::unique_ptr<DepNode>> &deps) {
	std::string path = dir + "/" + name;
	sys::FileInfo info = sys::fileInfo(path);

	if (info.isDir) {
		// May or may not need to copy variables
		BBVariables subvars;
		const BBVariables *varsptr = &variables;
		std::vector<std::string> subpaths;

		// There might be a $path/build.bb. If there is, files in that dir
		// needs a new copy of variables, and that $path/build.bb might contain
		// a files property.
		if (sys::fileExists(path + "/build.bb")) {
			subvars = variables;
			varsptr = &subvars;

			std::ifstream stream("build.bb");
			BBParser parser(stream, BBParser::FLAG_NONE);
			parser.parse(subvars);

			auto it = subvars.find("files");
			if (it == subvars.end()) {
				sys::readDir(path, subpaths);
			} else {
				subpaths = it->second;
				subvars.erase(it);
			}
		} else {
			sys::readDir(path, subpaths);
		}

		std::sort(subpaths.begin(), subpaths.end());
		for (auto &subpath: subpaths) {
			findDeps(path, subpath, *varsptr, deps);
		}
	} else {
		addFile(std::move(path), variables, deps);
	}
}

static std::string findTargetName(const BBVariables &variables) {
	auto it = variables.find("target");
	if (it != variables.end() && it->second.size() != 0) {
		return it->second[0];
	}

	// TODO: Find from deps

	return "target";
}

std::unique_ptr<DepNode> buildDepTree(const std::string &outDir, BBVariables variables) {
	// Read config from file
	if (sys::fileExists("build.bb")) {
		std::ifstream stream("build.bb");
		BBParser parser(stream, BBParser::FLAG_NONE);
		parser.parse(variables);
	}

	// Find source files/dirs
	std::vector<std::string> sourcePaths;
	{
		auto it = variables.find("files");
		if (it == variables.end()) {
			sys::readDir(".", sourcePaths);
		} else {
			sourcePaths = it->second;
			variables.erase(it);
		}
	}

	// Find dependencies
	std::vector<std::unique_ptr<DepNode>> deps;
	for (std::string &path: sourcePaths) {
		findDeps(".", path, variables, deps);
	}

	std::unique_ptr<LinkStep> link = std::make_unique<LinkStep>(
		findTargetName(variables), toolchain::TargetType::BINARY);
	for (auto &dep: deps) {
		link->addChild(std::move(dep));
	}

	return link;
}
