#include "parallel.h"

#include <thread>
#include <condition_variable>
#include <memory>
#include <vector>

namespace parallel {

struct Worker {
	bool running;
	std::function<void(void)> func;
	std::condition_variable cond;
	std::mutex mut;

	std::thread thread;
	Worker *next_worker;
};

static std::condition_variable workers_cond;
static std::mutex workers_mut;
static std::unique_ptr<Worker[]> workers;
static Worker *head= nullptr;
static int threads_running = 0;
static int waiting = 0;

static void workerFunc(Worker *w) {
	std::unique_lock<std::mutex> lock(w->mut);
	while (true) {
		w->cond.wait(lock, [w] { return (bool)w->func || !w->running; });
		if (!w->running)
			break;

		try {
			w->func();
		} catch (...) {
			std::terminate();
		}

		w->func = nullptr;
		std::unique_lock<std::mutex> workers_lock(workers_mut);
		w->next_worker = head;
		head = w;
		workers_lock.unlock();
		workers_cond.notify_one();
		threads_running -= 1;
	}
}

ParallelContext init(int num) {
	std::unique_lock<std::mutex> lock(workers_mut);
	if (num <= 1) {
		return ParallelContext{};
	}

	workers.reset(new Worker[num]);
	for (int i = 0; i < num; ++i) {
		Worker *w = &workers[i];
		w->next_worker = head;
		head = w;

		w->running = true;
		w->thread = std::thread(workerFunc, w);
	}

	return ParallelContext{};
}

ParallelContext::~ParallelContext() {
	std::unique_lock<std::mutex> lock(workers_mut);
	workers_cond.wait(lock, [] { return threads_running == 0 && waiting == 0; });

	Worker *w = head;
	while (w) {
		w->running = false;
		w->cond.notify_one();
		w = w->next_worker;
	}

	w = head;
	while (w) {
		w->thread.join();
		w = w->next_worker;
	}

	workers.reset();
	head = nullptr;
}

void run(std::function<void(void)> func) {
	std::unique_lock<std::mutex> lock(workers_mut);
	if (!workers) {
		func();
		return;
	}

	waiting += 1;
	workers_cond.wait(lock, [] { return head != nullptr; });
	Worker *w = head;
	head = head->next_worker;
	threads_running += 1;
	waiting -= 1;
	lock.unlock();

	// Launch!
	w->func = std::move(func);
	w->cond.notify_one();
}

int coreCount() {
	return std::thread::hardware_concurrency();
}

}
